#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gtk/gtk.h>
#include "specimen.h"
#include "patch.h"
#include "mixer.h"

/* gui specific headers */
#include "gui.h"
#include "waveform.h"
#include "sample-editor.h"
#include "sample-selector.h"
#include "bank-ops.h"
#include "piano.h"
#include "bank-builder.h"
#include "audio-settings.h"
#include "playmode-settings.h"
#include "filter-settings.h"
#include "param-settings.h"
#include "porta-settings.h"

/* parameter button codes for cb_setting_clicked() */
enum
{
     SETTING_FILTER,
     SETTING_MODE,
     SETTING_PARAMS,
     SETTING_PORTA,
};

/* main window */
static GtkWidget *window;
static GtkWidget *window_vbox;
static GtkWidget *master_hbox;
static GtkWidget *left_vbox;
static GtkWidget *right_vbox;

/* controls */
static GtkWidget *master_vol_spin;
static GtkWidget *chan_spin;
static GtkWidget *note_spin;
static GtkWidget *vol_spin;
static GtkWidget *pan_spin;
static GtkWidget *file_button;
static GtkWidget *file_label;
static GtkWidget *play_button;
static GtkWidget *patch_option_menu;
static GtkWidget *pitch_spin;
static GtkWidget *pitch_steps_spin;
static GtkWidget *cut_spin;
static GtkWidget *cut_by_spin;
static GtkWidget *range_check;
static GtkWidget *range_low_spin;
static GtkWidget *range_up_spin;
static GtkWidget *waveform;

GtkWidget *bb;
Piano *piano;

/* main menu */
static GtkWidget *menu_file;
static GtkWidget *menu_file_item;
static GtkWidget *menu_file_new_bank;
static GtkWidget *menu_file_open_bank;
static GtkWidget *menu_file_save_bank;
static GtkWidget *menu_file_save_bank_as;
static GtkWidget *menu_file_quit;
static GtkWidget *menu_file_vsep;
static GtkWidget *menu_settings;
static GtkWidget *menu_settings_item;
static GtkWidget *menu_settings_audio;

/* global adjustments */
static GtkAdjustment *master_vol_adj;
static GtkAdjustment *chan_adj;
static GtkAdjustment *note_adj;
static GtkAdjustment *vol_adj;
static GtkAdjustment *pan_adj;
static GtkAdjustment *pitch_adj;
static GtkAdjustment *pitch_steps_adj;
static GtkAdjustment *cut_adj;
static GtkAdjustment *cut_by_adj;
static GtkAdjustment *range_low_adj;
static GtkAdjustment *range_up_adj;

/* forward declarations */
static void set_current_patch (int id);

/* refreshes list of current patches */
static int update_interface (int index)
{
     static int count = 0;
     static GtkWidget **item = NULL;
     static GtkWidget *menu = NULL;
     int *patch;			/* array of active patches gets dumped here */
     int i;
     char *name;

     /* destroy any menu items */
     for (i = 0; i < count; i++)
	  if (item[i] != NULL)
	       gtk_widget_destroy (item[i]);

     /* destroy menu */
     if (menu != NULL)
	  gtk_widget_destroy (menu);

     /* create new menu */
     menu = gtk_menu_new ( );

     /* get array of active patches */
     count = patch_dump (&patch);
     if (count < 0)
	  debug
	       ("Failed to update interface, couldn't get array of active patches!\n");

     if (count == 0)
     {
	  /* better to attach an empty menu than no menu at all */
	  gtk_option_menu_set_menu (GTK_OPTION_MENU (patch_option_menu),
				    menu);

	  gtk_label_set_text (GTK_LABEL (file_label), "Load Sample");

	  set_current_patch (-1);

	  return 0;
     }

     /* create array of menu items */
     item = malloc (sizeof (GtkMenuItem *) * count);
     if (item == NULL)
     {
	  errmsg ("Failed to allocate space for patch menu!\n");
	  exit (FUBAR);
     }

     /* create a menu item for each patch id with appropriate callback */
     for (i = 0; i < count; i++)
     {
	  if ((name = patch_get_name (patch[i])) == NULL)
	  {
	       errmsg ("Couldn't find name for patch %d.\n", patch[i]);
	       exit (FUBAR);
	  }
	  item[i] = gtk_menu_item_new_with_label (name);
	  gtk_menu_shell_append (GTK_MENU_SHELL (menu), item[i]);
	  g_signal_connect_swapped (G_OBJECT (item[i]), "activate",
				    G_CALLBACK (set_current_patch),
				    (gpointer) patch[i]);
	  gtk_widget_show (item[i]);
	  if (name != NULL)
	       free (name);
     }

     /* free patch array */
     if (patch != NULL)
	  free (patch);

     /* attach new menu to patch option-menu */
     gtk_option_menu_set_menu (GTK_OPTION_MENU (patch_option_menu), menu);

     /* activate the requested menu item */
     g_signal_emit_by_name (G_OBJECT (item[index]), "activate");
     gtk_option_menu_set_history (GTK_OPTION_MENU (patch_option_menu),
				  index);

     return count;
}

/* A helper function so we can update the GUI from other actions */
int update_interface_helper (int index)
{
     return update_interface (index);
}

/* set the low range spinner */
void set_range_low_spin (int note)
{
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (range_low_spin), note);
}

/* set the upper range spinner */
void set_range_up_spin (int note)
{
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (range_up_spin), note);
}

void set_note_spin (int note)
{
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (note_spin), note);
}

void set_range_active ( )
{
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (range_check), 1);
}

/* when the 'x' button is clicked on the titlebar */
static gboolean cb_delete (GtkWidget* widget, GdkEvent* event,
			   gpointer data)
{
     /* we return false, causing the "destroy" event to be released on "widget" */
     return FALSE;
}

static void cb_quit (GtkWidget* widget, gpointer data)
{
     debug ("Quitting\n");
     gtk_main_quit ( );
}

static void cb_master_vol_set (GtkWidget* widget, GtkSpinButton* spin)
{
     float volume;
     int val;

     volume = gtk_spin_button_get_value_as_float (GTK_SPIN_BUTTON (spin));
     val = mixer_set_volume (volume / 100.0);
     if (val < 0)
	  errmsg ("Failed to set master volume to %.3f\n", volume);

     debug ("Master volume set to %.3f\n", volume);
}

static void cb_chan_set (GtkWidget * widget, GtkSpinButton * spin)
{
     int channel;
     int val;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     channel = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (spin));
     if ((val = patch_set_channel (cp, channel - 1)) < 0)
     {
	  errmsg ("Failed to set patch %d to channel %d (%s)\n", cp, val,
		  patch_strerror (channel));
	  return;
     }
     debug ("Patch %d channel set to %d\n", cp, channel - 1);
}

static void cb_note_set (GtkWidget * widget, GtkSpinButton * spin)
{
     int note;
     int val;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     note = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (spin));
     if ((val = patch_set_note (cp, note)) < 0)
     {
	  errmsg ("Failed to set patch %d to note %d (%s)\n", cp, note,
		  patch_strerror (val));
     }
     debug ("Patch %d now listening on note %d\n", cp, note);
     /* Show note on piano */
     piano_draw_range (piano);

     /* make sure range selection widgets can't do stupid stuff */
     gtk_spin_button_set_range (GTK_SPIN_BUTTON (range_low_spin), 0, note);
     gtk_spin_button_set_range (GTK_SPIN_BUTTON (range_up_spin), note, 127);
}

static void cb_range_set (GtkWidget * button, gpointer data)
{
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button)))
     {
	  gtk_widget_set_sensitive (range_low_spin, TRUE);
	  gtk_widget_set_sensitive (range_up_spin, TRUE);
	  patch_set_range (cp, 1);
	  piano_draw_range (piano);
	  debug ("Patch %d now listening to a range of notes\n", cp);
     }
     else
     {
	  gtk_widget_set_sensitive (range_low_spin, FALSE);
	  gtk_widget_set_sensitive (range_up_spin, FALSE);
	  patch_set_range (cp, 0);
	  piano_draw_range (piano);
	  debug ("Patch %d note listening to an individual note\n", cp);
     }
}

static void cb_lower_note_set (GtkWidget * widget, gpointer data)
{
     int note, val;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     note = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (widget));
     if ((val = patch_set_lower_note (cp, note)) < 0)
     {
	  debug ("Unable to set patch %d lower note to %d (%s)\n", cp, note,
		 patch_strerror (val));
	  return;
     }
     piano_draw_range (piano);
     debug ("Patch %d lower note set to %d\n", cp, note);
}

static void cb_upper_note_set (GtkWidget * widget, gpointer data)
{
     int note, val;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     note = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (widget));
     if ((val = patch_set_upper_note (cp, note)) < 0)
     {
	  debug ("Unable to set patch %d upper note to %d (%s)\n", cp, note,
		 patch_strerror (val));
	  return;
     }
     piano_draw_range (piano);
     debug ("Patch %d upper note set to %d\n", cp, note);
}

static void cb_vol_set (GtkWidget * widget, gpointer data)
{
     float volume;
     int val;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     volume = gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget));
     volume /= 100;
     if ((val = patch_set_volume (cp, volume)) < 0)
     {
	  debug ("Unable to set patch %d volume to %.3f (%s)\n", cp, volume,
		 patch_strerror (val));
	  return;
     }
     debug ("Patch %d volume set to %.3f\n", cp, volume);
}

static void cb_pan_set (GtkWidget * widget, gpointer data)
{
     float pan;
     int val;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     pan = gtk_spin_button_get_value (GTK_SPIN_BUTTON (widget));
     pan /= 100;
     if ((val = patch_set_panning (cp, pan)) < 0)
     {
	  debug ("Unable to set patch %d panning to %.3f (%s)\n", cp, pan,
		 patch_strerror (val));
	  return;
     }
     debug ("Patch %d pan set to %.3f\n", cp, pan);
}

static void cb_pitch_set (GtkWidget* widget, gpointer data)
{
     float pitch;
     int val;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     pitch = gtk_spin_button_get_value (GTK_SPIN_BUTTON (pitch_spin));
     if ((val = patch_set_pitch (cp, pitch)) < 0)
     {
	  debug ("Unable to set patch %d pitch to %f (%s)\n", cp, pitch,
		 patch_strerror (val));
	  return;
     }
}

static void cb_pitch_steps_set (GtkWidget* widget, gpointer data)
{
     int steps;
     int val;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     steps = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (pitch_steps_spin));
     if ((val = patch_set_pitch_steps (cp, steps)) < 0)
     {
	  debug ("Unable to set patch %d pitch steps to %d (%s)\n", cp, steps,
		 patch_strerror (val));
	  return;
     }
}

static void cb_cut_set (GtkWidget * widget, gpointer arg)
{
     int val;
     int cut;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     cut = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (cut_spin));
     if ((val = patch_set_cut (cp, cut)) < 0)
     {
	  debug ("Unable to set patch %d cut signal to %d (%s)\n", cp, cut,
		 patch_strerror (val));
	  return;
     }
     debug ("Patch %d cut signal set to %d\n", cp, cut);
}

static void cb_cut_by_set (GtkWidget * widget, gpointer arg)
{
     int val;
     int cut_by;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     cut_by =
	  gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (cut_by_spin));
     if ((val = patch_set_cut_by (cp, cut_by)) < 0)
     {
	  debug ("Unable to set patch %d cut_by signal to %d (%s)\n", cp,
		 cut_by, patch_strerror (val));
	  return;
     }
     debug ("Patch %d cut_by signal set to %d\n", cp, cut_by);
}

/* enables ok button if there is input in entry, disables otherwise */
static gboolean cb_action_name_verify (GtkWidget * entry,
				       GdkEventKey * event,
				       GtkDialog * dialog)
{
     int val = strlen ((char *) gtk_entry_get_text (GTK_ENTRY (entry)));

     if (val)
	  gtk_dialog_set_response_sensitive (dialog, GTK_RESPONSE_ACCEPT,
					     TRUE);
     else
	  gtk_dialog_set_response_sensitive (dialog, GTK_RESPONSE_ACCEPT,
					     FALSE);

     return FALSE;		/* make sure normal event handlers grab signal */
}

static void cb_action_add (GtkWidget * menu_item, GtkWidget * main_window)
{
     GtkWidget *dialog;
     GtkWidget *entry;
     int val;
     int index;

     /* dialog box */
     dialog = gtk_dialog_new_with_buttons ("Add Patch",
					   GTK_WINDOW (main_window),
					   GTK_DIALOG_MODAL |
					   GTK_DIALOG_DESTROY_WITH_PARENT,
					   GTK_STOCK_OK,
					   GTK_RESPONSE_ACCEPT,
					   GTK_STOCK_CANCEL,
					   GTK_RESPONSE_REJECT, NULL);
     gtk_dialog_set_default_response (GTK_DIALOG (dialog),
				      GTK_RESPONSE_ACCEPT);

     /* create entry box */
     entry = gtk_entry_new_with_max_length (PATCH_MAX_NAME);
     gtk_entry_set_text (GTK_ENTRY (entry), "Patch Name");
     gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);

     /* I wish I had a "changed" event... */
     g_signal_connect (G_OBJECT (entry), "key-press-event",
		       G_CALLBACK (cb_action_name_verify),
		       (gpointer) dialog);
     g_signal_connect (G_OBJECT (entry), "key-release-event",
		       G_CALLBACK (cb_action_name_verify),
		       (gpointer) dialog);
     gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), entry, FALSE,
			 FALSE, 0);
     gtk_widget_show (entry);

     /* guaranteed to have a string in there if response is 'accept'
      * due to sensitivity callback */
     if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
     {
	  val =
	       patch_create ((char *) gtk_entry_get_text (GTK_ENTRY (entry)));
	  if (val < 0)
	  {
	       errmsg ("Failed to create a new patch (%s).\n",
		       patch_strerror (val));
	       return;
	  }
	  index = patch_get_display_index (val);

	  update_interface (index);
     }

     gtk_widget_destroy (dialog);
}

static void cb_action_duplicate (GtkWidget * menu_item,
				 GtkWidget * main_window)
{
     int val, index;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
     {
	  debug ("Duplicate what, jackass?\n");
	  return;
     }

     val = patch_duplicate (cp);
     if (val < 0)
     {
	  errmsg ("Failed to create a new patch (%s).\n",
		  patch_strerror (val));
	  return;
     }
     index = patch_get_display_index (val);

     update_interface (index);
}

static void cb_action_rename (GtkWidget * menu_item,
			      GtkWidget * main_window)
{
     GtkWidget *dialog;
     GtkWidget *entry;
     int val;
     int index;
     int cp;

     if ((cp = get_current_patch ( )) < 0)
     {
	  debug ("No patches to rename, infidel.\n");
	  return;
     }

     /* dialog box */
     dialog = gtk_dialog_new_with_buttons ("Rename Patch",
					   GTK_WINDOW (main_window),
					   GTK_DIALOG_MODAL |
					   GTK_DIALOG_DESTROY_WITH_PARENT,
					   GTK_STOCK_OK,
					   GTK_RESPONSE_ACCEPT,
					   GTK_STOCK_CANCEL,
					   GTK_RESPONSE_REJECT, NULL);
     gtk_dialog_set_default_response (GTK_DIALOG (dialog),
				      GTK_RESPONSE_ACCEPT);

     /* create entry box */
     entry = gtk_entry_new_with_max_length (PATCH_MAX_NAME);
     gtk_entry_set_text (GTK_ENTRY (entry), patch_get_name (cp));
     gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);

     /* I wish I had a "changed" event... */
     g_signal_connect (G_OBJECT (entry), "key-press-event",
		       G_CALLBACK (cb_action_name_verify),
		       (gpointer) dialog);
     g_signal_connect (G_OBJECT (entry), "key-release-event",
		       G_CALLBACK (cb_action_name_verify),
		       (gpointer) dialog);
     gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), entry, FALSE,
			 FALSE, 0);
     gtk_widget_show (entry);

     /* guaranteed to have a string in there if response is 'accept'
      * due to sensitivity callback */
     if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
     {
	  val =
	       patch_set_name (cp,
			       (char *)
			       gtk_entry_get_text (GTK_ENTRY (entry)));
	  if (val < 0)
	  {
	       errmsg ("Failed to rename patch (%s).\n",
		       patch_strerror (val));
	       return;
	  }

	  index = patch_get_display_index (cp);
	  update_interface (index);
     }

     gtk_widget_destroy (dialog);
}

static void cb_action_remove (GtkWidget * menu_item, gpointer data)
{
     int val;
     int cp;
     int index;

     if ((cp = get_current_patch ( )) < 0)
     {
	  debug ("No patches to remove, you dolt.\n");
	  return;
     }

     index = patch_get_display_index (cp);
     if ((val = patch_destroy (cp)) < 0)
     {
	  errmsg ("Error removing patch %d (%s).\n", cp,
		  patch_strerror (val));
	  return;
     }

     if (index == 0)
	  update_interface (index);
     else
	  update_interface (index - 1);
}

/* ripped off from gtk-2.0 tutorial, makes menu popup when button is clicked */
static gboolean cb_action_popup (GtkWidget * menu, GdkEvent * event)
{
     GdkEventButton *bevent;

     if (event->type == GDK_BUTTON_PRESS)
     {
	  bevent = (GdkEventButton *) event;
	  gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
			  bevent->button, bevent->time);
	  /* Tell calling code that we have handled this event; the buck
	   * stops here. */
	  return TRUE;
     }

     /* Tell calling code that we have not handled this event; pass it on. */
     return FALSE;
}

static void cb_sample_load (GtkWidget * button, gpointer data)
{
     int cp;
     int index;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     sample_selector_show ( );
     index = patch_get_display_index (cp);
     update_interface (index);
}

/* Drag and drop to load a new sample */
void cb_sample_dnd (GtkWidget * widget, GdkDragContext * context,
		    gint x, gint y, GtkSelectionData * selection_data,
		    guint info, guint time, gpointer data)
{
     GtkWidget *msg;
     char *name = NULL;
     char *curname = NULL;
     int cp, val;
     int index;

     debug ("we got a drop! \n");
     GSList *filelist = NULL;

     if (!selection_data->data)
     {
	  return;
     }

     /* get filename then create new patch, selection data is actually the filenames all as text,
      * seperated by new lines so if we want multifile support we have to process that... */
     /* the file selector gives us a uri, so we have to trim the file:/// stuff off it. */
     filelist = uri_list_to_gslist (selection_data->data);

     if ((cp = get_current_patch ( )) < 0)
	  return;

     name = g_slist_nth_data (filelist, 0);
     curname = patch_get_sample_name (cp);

     /* don't load the sample if it's already loaded */
     if (strcmp (name, curname) == 0)
     {
	  if (curname != NULL)
	       free (curname);
	  return;
     }

     if ((val = patch_sample_load (cp, name)) < 0)
     {
	  errmsg ("Failed to load sample %s for patch %d (%s)\n", name, cp,
		  patch_strerror (val));
	  msg =
	       gtk_message_dialog_new (GTK_WINDOW (window), GTK_DIALOG_MODAL,
				       GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
				       "Failed to load sample %s", name);
	  gtk_dialog_run (GTK_DIALOG (msg));
	  gtk_widget_destroy (msg);
     }
     else
     {
	  debug ("Successfully loaded sample %s\n", name);
     }

     if (curname != NULL)
	  free (curname);

     index = patch_get_display_index (cp);
     update_interface (index);

}

static void cb_menu_file_new_bank (GtkWidget * widget, gpointer data)
{
     if (bank_ops_new ( ) == 0)
	  update_interface (0);
}

static void cb_menu_file_open_bank (GtkWidget * widget, gpointer data)
{
     if (bank_ops_open ( ) == 0)
	  update_interface (0);
}

static void cb_menu_file_save_bank (GtkWidget * widget, gpointer data)
{
     bank_ops_save ( );
}

static void cb_menu_file_save_bank_as (GtkWidget * widget, gpointer data)
{
     bank_ops_save_as ( );
}

static void cb_menu_settings_audio (GtkWidget * widget, gpointer data)
{
     audio_settings_show (window);
}

/* this is my favorite function */
static gboolean cb_play (GtkWidget * widget, GdkEvent * event,
			 gpointer data)
{
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return FALSE;

     if (event->type == GDK_BUTTON_PRESS || event->type == GDK_KEY_PRESS)
	  mixer_note_on_with_id (cp, patch_get_note (cp), 127);
     else if (event->type == GDK_BUTTON_RELEASE
	      || event->type == GDK_KEY_RELEASE)
	  mixer_note_off_with_id (cp, patch_get_note (cp));

     return FALSE;
}

static gboolean cb_waveform_button_press (GtkWidget * wf, GdkEvent * event,
					  gpointer data)
{
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return FALSE;

     if (event->type == GDK_BUTTON_PRESS)
	  sample_editor_show ( );

     /* this doesn't work, I need to get the waveform to refresh
      * whenever it's parent toplevel is idle */
     gtk_widget_queue_draw (wf);
     return FALSE;
}

static void cb_setting_clicked (GtkWidget* widget, int setting)
{
     int cp;

     if ((cp = get_current_patch ( )) < 0)
	  return;

     switch (setting)
     {
     case SETTING_FILTER:
	  filter_settings_show ( );
	  break;
     case SETTING_MODE:
	  playmode_settings_show ( );
	  break;
     case SETTING_PARAMS:
	  param_settings_show ( );
	  break;
     case SETTING_PORTA:
	  porta_settings_show ( );
	  break;
     }

     return;
}

int gui_init ( )
{
     /* frame stuff */
     GtkWidget *frame;
     GtkWidget *frame_vbox;
     GtkWidget *frame_hbox;
     GtkWidget *table;

     /* misc */
     GtkWidget *button;
     GtkWidget *hbox;
     GtkWidget *label;
     GtkWidget *vsep;
     GtkWidget *image;
     GtkWidget *menubar;
     GtkWidget *menu;
     GtkWidget *item;
     GtkWidget *top_canvas;
     GtkWidget *scrollwin;

     static const GtkTargetEntry targets[] = {
	  {"text/uri-list", 0, 0},
	  {"text/plain", 0, 0}
     };

     debug ("Initializing GUI\n");

     /****************************MAIN WINDOW****************************************/
     window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
     gtk_window_set_title (GTK_WINDOW (window), "Specimen");
     gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
     g_signal_connect (G_OBJECT (window), "delete-event",
		       G_CALLBACK (cb_delete), NULL);
     g_signal_connect (G_OBJECT (window), "destroy", G_CALLBACK (cb_quit),
		       NULL);

     /* setup the window's main vbox */
     window_vbox = gtk_vbox_new (FALSE, 0);
     gtk_container_add (GTK_CONTAINER (window), window_vbox);

     /* the menubar */
     menubar = gtk_menu_bar_new ( );
     gtk_box_pack_start (GTK_BOX (window_vbox), menubar, FALSE, FALSE, 0);
     gtk_widget_show (menubar);

     /* file menu */
     menu_file = gtk_menu_new ( );
     menu_file_item = gtk_menu_item_new_with_label ("File");
     gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_file_item), menu_file);
     gtk_menu_bar_append (GTK_MENU_BAR (menubar), menu_file_item);
     gtk_widget_show (menu_file_item);

     menu_file_new_bank = gtk_menu_item_new_with_label ("New Bank");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu_file), menu_file_new_bank);
     g_signal_connect (G_OBJECT (menu_file_new_bank), "activate",
		       G_CALLBACK (cb_menu_file_new_bank), NULL);
     gtk_widget_show (menu_file_new_bank);

     menu_file_open_bank = gtk_menu_item_new_with_label ("Open Bank...");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu_file),
			    menu_file_open_bank);
     g_signal_connect (G_OBJECT (menu_file_open_bank), "activate",
		       G_CALLBACK (cb_menu_file_open_bank), NULL);
     gtk_widget_show (menu_file_open_bank);

     menu_file_save_bank = gtk_menu_item_new_with_label ("Save Bank");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu_file),
			    menu_file_save_bank);
     g_signal_connect (G_OBJECT (menu_file_save_bank), "activate",
		       G_CALLBACK (cb_menu_file_save_bank), NULL);
     gtk_widget_show (menu_file_save_bank);

     menu_file_save_bank_as =
	  gtk_menu_item_new_with_label ("Save Bank As...");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu_file),
			    menu_file_save_bank_as);
     g_signal_connect (G_OBJECT (menu_file_save_bank_as), "activate",
		       G_CALLBACK (cb_menu_file_save_bank_as), NULL);
     gtk_widget_show (menu_file_save_bank_as);

     menu_file_vsep = gtk_menu_item_new ( );
     gtk_menu_shell_append (GTK_MENU_SHELL (menu_file), menu_file_vsep);
     gtk_widget_show (menu_file_vsep);

     menu_file_quit = gtk_menu_item_new_with_label ("Quit");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu_file), menu_file_quit);
     g_signal_connect (G_OBJECT (menu_file_quit), "activate",
		       G_CALLBACK (cb_quit), NULL);
     gtk_widget_show (menu_file_quit);

     /* settings menu */
     menu_settings = gtk_menu_new ( );
     menu_settings_item = gtk_menu_item_new_with_label ("Settings");
     gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_settings_item),
				menu_settings);
     gtk_menu_bar_append (GTK_MENU_BAR (menubar), menu_settings_item);
     gtk_widget_show (menu_settings_item);

     menu_settings_audio = gtk_menu_item_new_with_label ("Audio");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu_settings),
			    menu_settings_audio);
     g_signal_connect (G_OBJECT (menu_settings_audio), "activate",
		       G_CALLBACK (cb_menu_settings_audio), NULL);
     gtk_widget_show (menu_settings_audio);

     /* setup the main window's master hbox, and left and right boxes */
     master_hbox = gtk_hbox_new (FALSE, SPACING);
     left_vbox = gtk_vbox_new (FALSE, SPACING);
     right_vbox = gtk_vbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (window_vbox), master_hbox, FALSE, FALSE,
			 0);
     gtk_box_pack_start (GTK_BOX (master_hbox), left_vbox, FALSE, FALSE, 0);
     gtk_box_pack_start (GTK_BOX (master_hbox), right_vbox, FALSE, FALSE,
			 0);
     gtk_container_set_border_width (GTK_CONTAINER (left_vbox), SPACING);
     gtk_container_set_border_width (GTK_CONTAINER (right_vbox), SPACING);

     /****************************MASTER FRAME***************************************/
     frame = gtk_frame_new ("Master");
     gtk_box_pack_start (GTK_BOX (left_vbox), frame, FALSE, FALSE, 0);
     gtk_widget_show (frame);

     /* master table */
     table = gtk_table_new (2, 2, FALSE);
     gtk_container_set_border_width (GTK_CONTAINER (table), SPACING);
     gtk_table_set_row_spacings (GTK_TABLE (table), SPACING);
     gtk_table_set_col_spacings (GTK_TABLE (table), SPACING);
     gtk_container_add (GTK_CONTAINER (frame), table);
     gtk_widget_show (table);

     /*---------------------------ROW ONE-------------------------------------------*/
     /*====COL ONE====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 0, 1, 0, 1);
     gtk_widget_show (frame_hbox);

     /* patch option-menu */
     patch_option_menu = gtk_option_menu_new ( );
     gtk_box_pack_end (GTK_BOX (frame_hbox), patch_option_menu, TRUE, TRUE,
		       0);
     gtk_widget_show (patch_option_menu);

     /* patch option-menu label */
     label = gtk_label_new ("Patch:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /*====COL TWO====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 1, 2, 0, 1);
     gtk_widget_show (frame_hbox);

     /* patch option-menu action button menu */
     menu = gtk_menu_new ( );

     /*-add-*/
     item = gtk_menu_item_new_with_label ("Add");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_action_add), (gpointer) window);
     gtk_widget_show (item);

     /*-duplicate-*/
     item = gtk_menu_item_new_with_label ("Duplicate");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_action_duplicate), (gpointer) window);
     gtk_widget_show (item);

     /*-rename-*/
     item = gtk_menu_item_new_with_label ("Rename");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_action_rename), (gpointer) window);
     gtk_widget_show (item);

     /*-remove-*/
     item = gtk_menu_item_new_with_label ("Remove");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_action_remove), (gpointer) NULL);
     gtk_widget_show (item);

     /* patch option-menu action button */
     button = gtk_button_new_with_label ("Action");
     gtk_box_pack_end (GTK_BOX (frame_hbox), button, FALSE, FALSE, 0);
     g_signal_connect_swapped (G_OBJECT (button), "event",
			       G_CALLBACK (cb_action_popup),
			       G_OBJECT (menu));
     gtk_widget_show (button);

     /*---------------------------ROW TWO-------------------------------------------*/
     /*====COL ONE====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 0, 1, 1, 2);
     gtk_widget_show (frame_hbox);

     /* master volume spinbutton adjustment */
     master_vol_adj =
	  (GtkAdjustment *) gtk_adjustment_new (70, 0, 100, 1, 0, 0);

     /* master volume spinbutton */
     master_vol_spin = gtk_spin_button_new (master_vol_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (frame_hbox), master_vol_spin, FALSE, FALSE,
		       0);
     gtk_widget_show (master_vol_spin);
     g_signal_connect (G_OBJECT (master_vol_adj), "value-changed",
		       G_CALLBACK (cb_master_vol_set),
		       (gpointer) master_vol_spin);
     g_signal_emit_by_name (G_OBJECT (master_vol_adj), "value-changed");

     /* volume label */
     label = gtk_label_new ("Volume:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /***************************SAMPLE FRAME***********************************/
     frame = gtk_frame_new ("Sample");
     gtk_box_pack_start (GTK_BOX (left_vbox), frame, FALSE, FALSE, 0);
     gtk_widget_show (frame);

     /* master table */
     table = gtk_table_new (5, 2, FALSE);
     gtk_container_set_border_width (GTK_CONTAINER (table), SPACING);
     gtk_table_set_row_spacings (GTK_TABLE (table), SPACING);
     gtk_table_set_col_spacings (GTK_TABLE (table), SPACING);
     gtk_container_add (GTK_CONTAINER (frame), table);
     gtk_widget_show (table);

     /*----------------------------ROW ONE----------------------------------------*/
     /*====COL ONE====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 0, 2, 0, 1);
     gtk_widget_show (frame_hbox);

     /* play button */
     play_button = gtk_button_new_with_label ("Play");
     gtk_box_pack_end (GTK_BOX (frame_hbox), play_button, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (play_button), "event",
		       G_CALLBACK (cb_play), NULL);
     gtk_widget_show (play_button);

     /* file label */
     label = gtk_label_new ("File:");
     gtk_box_pack_start (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* file browsing button */
     file_button = gtk_button_new ( );
     hbox = gtk_hbox_new (FALSE, SPACING);
     image = gtk_image_new_from_file (PIXMAPSDIR "open.png");
     file_label = gtk_label_new ("Load Sample");
     vsep = gtk_vseparator_new ( );

     gtk_box_pack_start (GTK_BOX (hbox), file_label, FALSE, FALSE, 0);
     gtk_box_pack_start (GTK_BOX (hbox), vsep, FALSE, FALSE, 0);
     gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
     gtk_container_add (GTK_CONTAINER (file_button), hbox);
     gtk_box_pack_start (GTK_BOX (frame_hbox), file_button, FALSE, FALSE,
			 0);

     gtk_drag_dest_set (file_button, GTK_DEST_DEFAULT_ALL, targets, 1,
			GDK_ACTION_COPY | GDK_ACTION_MOVE);
     g_signal_connect (G_OBJECT (file_button), "drag_data_received",
		       G_CALLBACK (cb_sample_dnd), NULL);

     g_signal_connect (G_OBJECT (file_button), "clicked",
		       G_CALLBACK (cb_sample_load), NULL);

     gtk_widget_show (file_label);
     gtk_widget_show (vsep);
     gtk_widget_show (image);
     gtk_widget_show (hbox);
     gtk_widget_show (file_button);

     /*----------------------------ROW TWO----------------------------------------*/
     /*====COL ONE====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 0, 1, 1, 2);
     gtk_widget_show (frame_hbox);

     /* volume spin adjustment */
     vol_adj = (GtkAdjustment *) gtk_adjustment_new (100, 0, 100, 1, 0, 0);

     /* volume spin */
     vol_spin = gtk_spin_button_new (vol_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (frame_hbox), vol_spin, FALSE, FALSE, 0);
     g_signal_connect_swapped (G_OBJECT (vol_adj), "value-changed",
			       G_CALLBACK (cb_vol_set),
			       (gpointer) vol_spin);
     gtk_widget_show (vol_spin);

     /* volume label */
     label = gtk_label_new ("Volume:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /*====COL TWO====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 1, 2, 1, 2);
     gtk_widget_show (frame_hbox);

     /* pan spin adjustment */
     pan_adj = (GtkAdjustment *) gtk_adjustment_new (0, -100, 100, 1, 0, 0);

     /* pan spin button */
     pan_spin = gtk_spin_button_new (pan_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (frame_hbox), pan_spin, FALSE, FALSE, 0);
     g_signal_connect_swapped (G_OBJECT (pan_adj), "value-changed",
			       G_CALLBACK (cb_pan_set),
			       (gpointer) pan_spin);
     gtk_widget_show (pan_spin);

     /* pan label */
     label = gtk_label_new ("Pan:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /*----------------------------ROW THREE----------------------------------------*/
     /*====COL ONE====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 0, 1, 2, 3);
     gtk_widget_show (frame_hbox);

     /* pitch adjustment */
     pitch_adj = (GtkAdjustment*) gtk_adjustment_new (0.0, -1.0, 1.0, 0.01, 0.1, 0.0);

     /* pitch spin button */
     pitch_spin = gtk_spin_button_new (pitch_adj, 0.0, 2);
     gtk_box_pack_end (GTK_BOX (frame_hbox), pitch_spin, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (pitch_adj), "value-changed",
		       G_CALLBACK (cb_pitch_set), (gpointer) NULL);
     gtk_widget_show (pitch_spin);

     /* pitch label */
     label = gtk_label_new ("Pitch:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);
     
     /*====COL TWO====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 1, 2, 2, 3);
     gtk_widget_show (frame_hbox);

     /* pitch_steps adjustment */
     pitch_steps_adj = (GtkAdjustment*) gtk_adjustment_new (0, -PATCH_MAX_PITCH_STEPS,
							    PATCH_MAX_PITCH_STEPS, 1,
							    1, 0);

     /* pitch_steps spin button */
     pitch_steps_spin = gtk_spin_button_new (pitch_steps_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (frame_hbox), pitch_steps_spin, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (pitch_steps_adj), "value-changed",
		       G_CALLBACK (cb_pitch_steps_set), (gpointer) NULL);
     gtk_widget_show (pitch_steps_spin);

     /* pitch_steps label */
     label = gtk_label_new ("Steps:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /*----------------------------ROW FOUR----------------------------------------*/
     /*====COL ONE====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 0, 1, 3, 4);
     gtk_widget_show (frame_hbox);

     /* cut adjustment */
     cut_adj = (GtkAdjustment *) gtk_adjustment_new (0, 0, 99, 1, 0, 0);

     /* cut spin button */
     cut_spin = gtk_spin_button_new (cut_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (frame_hbox), cut_spin, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (cut_adj), "value-changed",
		       G_CALLBACK (cb_cut_set), (gpointer) NULL);
     gtk_widget_show (cut_spin);

     /* cut label */
     label = gtk_label_new ("Cut:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /*====COL TWO====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 1, 2, 3, 4);
     gtk_widget_show (frame_hbox);

     /* cutd by adjustment */
     cut_by_adj = (GtkAdjustment *) gtk_adjustment_new (0, 0, 99, 1, 0, 0);

     /* cutd by spin button */
     cut_by_spin = gtk_spin_button_new (cut_by_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (frame_hbox), cut_by_spin, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (cut_by_adj), "value-changed",
		       G_CALLBACK (cb_cut_by_set), (gpointer) NULL);
     gtk_widget_show (cut_by_spin);

     /* cutd by label */
     label = gtk_label_new ("Cut by:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /*----------------------------ROW FIVE----------------------------------------*/
     /*====COL ONE/TWO====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 0, 2, 4, 5);
     gtk_widget_show (frame_hbox);

     /* waveform preview */
     waveform = waveform_new (get_current_patch ( ), 256, 64, FALSE);
     gtk_box_pack_end (GTK_BOX (frame_hbox), waveform, TRUE, TRUE, 0);
     g_signal_connect (G_OBJECT (waveform), "button-press-event",
		       G_CALLBACK (cb_waveform_button_press), NULL);
     gtk_widget_show (waveform);

     /***************************MIDI FRAME***********************************/
     frame = gtk_frame_new ("Midi");

     /* put it in the right box if we don't have 2.4 */
#ifdef HAVE_GTK2_4
     gtk_box_pack_start (GTK_BOX (left_vbox), frame, FALSE, FALSE, 0);
#else
     gtk_box_pack_start (GTK_BOX (right_vbox), frame, FALSE, FALSE, 0);
#endif

     gtk_widget_show (frame);

     /* master table */
     table = gtk_table_new (2, 2, FALSE);
     gtk_container_set_border_width (GTK_CONTAINER (table), SPACING);
     gtk_table_set_row_spacings (GTK_TABLE (table), SPACING);
     gtk_table_set_col_spacings (GTK_TABLE (table), SPACING);
     gtk_container_add (GTK_CONTAINER (frame), table);
     gtk_widget_show (table);

     /*----------------------------ROW ONE----------------------------------------*/
     /*====COL ONE====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 0, 1, 0, 1);
     gtk_widget_show (frame_hbox);

     /* channel spinbutton adjustment */
     chan_adj = (GtkAdjustment *) gtk_adjustment_new (1, 1, 16, 1, 0, 0);

     /* channel spinbutton */
     chan_spin = gtk_spin_button_new (chan_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (frame_hbox), chan_spin, FALSE, FALSE, 0);
     gtk_widget_show (chan_spin);

     /* bind to proper callback */
     g_signal_connect (G_OBJECT (chan_adj), "value_changed",
		       G_CALLBACK (cb_chan_set), (gpointer) chan_spin);

     /* channel label */
     label = gtk_label_new ("Channel:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /*====COL TWO====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 1, 2, 0, 1);
     gtk_widget_show (frame_hbox);

     /* note spinbutton adjustment */
     note_adj = (GtkAdjustment *) gtk_adjustment_new (60, 0, 127, 1, 0, 0);

     /* note spinbutton */
     note_spin = gtk_spin_button_new (note_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (frame_hbox), note_spin, FALSE, FALSE, 0);
     gtk_widget_show (note_spin);

     /* bind to proper callback */
     g_signal_connect (G_OBJECT (note_adj), "value_changed",
		       G_CALLBACK (cb_note_set), (gpointer) note_spin);

     /* note label */
     label = gtk_label_new ("Note:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /*----------------------------ROW TWO----------------------------------------*/
     /*====COL ONE====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 0, 1, 1, 2);
     gtk_widget_show (frame_hbox);

     /* range lower spinbutton */
     range_low_adj =
	  (GtkAdjustment *) gtk_adjustment_new (60, 1, 127, 1, 0, 0);
     range_low_spin = gtk_spin_button_new (range_low_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (frame_hbox), range_low_spin, FALSE, FALSE,
		       0);
     g_signal_connect_swapped (G_OBJECT (range_low_adj), "value-changed",
			       G_CALLBACK (cb_lower_note_set),
			       (gpointer) range_low_spin);
     gtk_widget_show (range_low_spin);

     /* range check box */
     range_check = gtk_check_button_new_with_label ("Range from:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), range_check, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (range_check), "toggled",
		       G_CALLBACK (cb_range_set), NULL);
     gtk_widget_show (range_check);

     /*====COL TWO====*/
     frame_hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_table_attach_defaults (GTK_TABLE (table), frame_hbox, 1, 2, 1, 2);
     gtk_widget_show (frame_hbox);

     /* range upper spinbutton */
     range_up_adj =
	  (GtkAdjustment *) gtk_adjustment_new (60, 60, 127, 1, 0, 0);
     range_up_spin = gtk_spin_button_new (range_up_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (frame_hbox), range_up_spin, FALSE, FALSE,
		       0);
     g_signal_connect_swapped (G_OBJECT (range_up_adj), "value-changed",
			       G_CALLBACK (cb_upper_note_set),
			       (gpointer) range_up_spin);
     gtk_widget_show (range_up_spin);

     /* "to" label */
     label = gtk_label_new ("to:");
     gtk_box_pack_end (GTK_BOX (frame_hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);


     /****************************END LEFT BOX *************************/

     /************************* BEGIN RIGHT BOX *****************/

#ifdef HAVE_GTK2_4
     /* -------- Bank Builder frame ----------- */
     frame = gtk_frame_new ("Sample Chooser");
     gtk_box_pack_start (GTK_BOX (right_vbox), frame, TRUE, TRUE, 0);
     gtk_widget_show (frame);

     /* master vbox */
     frame_vbox = gtk_vbox_new (FALSE, SPACING);
     gtk_container_set_border_width (GTK_CONTAINER (frame_vbox), SPACING);
     gtk_container_add (GTK_CONTAINER (frame), frame_vbox);
     gtk_widget_show (frame_vbox);

     /*------------ROW ONE-------------------*/
     frame_hbox = gtk_hbox_new (TRUE, SPACING);
     gtk_box_pack_start (GTK_BOX (frame_vbox), frame_hbox, TRUE, TRUE, 0);
     gtk_widget_show (frame_hbox);
#endif
     /* Bank Builder, we still need an object even if we don't have gtk2.4 for DnD */
     bb = bankbuilder_new ( );
#ifdef HAVE_GTK2_4

     gtk_box_pack_start (GTK_BOX (frame_hbox), bb, TRUE, TRUE, 0);
     gtk_widget_show (bb);
#endif
     /***************SETTINGS FRAME***************/
     frame = gtk_frame_new ("Settings");
     gtk_box_pack_start (GTK_BOX (right_vbox), frame, FALSE, FALSE, 0);
     gtk_widget_show (frame);

     /* master vbox */
     frame_vbox = gtk_vbox_new (FALSE, SPACING);
     gtk_container_set_border_width (GTK_CONTAINER (frame_vbox), SPACING);
     gtk_container_add (GTK_CONTAINER (frame), frame_vbox);
     gtk_widget_show (frame_vbox);

     /*------------ROW ONE-------------------*/
     frame_hbox = gtk_hbox_new (TRUE, SPACING);
     gtk_box_pack_start (GTK_BOX (frame_vbox), frame_hbox, TRUE, TRUE, 0);
     gtk_widget_show (frame_hbox);

     /* mode button */
     button = gtk_button_new_with_label ("Mode");
     g_signal_connect (G_OBJECT (button), "clicked",
		       G_CALLBACK (cb_setting_clicked),
		       (gpointer) SETTING_MODE);
     gtk_box_pack_start (GTK_BOX (frame_hbox), button, TRUE, TRUE, 0);
     gtk_widget_show (button);

     /* filter button */
     button = gtk_button_new_with_label ("Filter");
     g_signal_connect (G_OBJECT (button), "clicked",
		       G_CALLBACK (cb_setting_clicked),
		       (gpointer) SETTING_FILTER);
     gtk_box_pack_start (GTK_BOX (frame_hbox), button, TRUE, TRUE, 0);
     gtk_widget_show (button);

     /* params button */
     button = gtk_button_new_with_label ("Params");
     g_signal_connect (G_OBJECT (button), "clicked",
		       G_CALLBACK (cb_setting_clicked),
		       (gpointer) SETTING_PARAMS);
     gtk_box_pack_start (GTK_BOX (frame_hbox), button, TRUE, TRUE, 0);
     gtk_widget_show (button);

     /* porta button */
     button = gtk_button_new_with_label ("Porta");
     g_signal_connect (G_OBJECT (button), "clicked",
		       G_CALLBACK (cb_setting_clicked),
		       (gpointer) SETTING_PORTA);
     gtk_box_pack_start (GTK_BOX (frame_hbox), button, TRUE, TRUE, 0);
     gtk_widget_show (button);

     /****************************END RIGHT BOX *************************/

     /****************************BEGIN BOTTOM BOX **********************/

     /***************PIANO FRAME***************/
     frame = gtk_frame_new ("Note range");
     gtk_box_pack_start (GTK_BOX (window_vbox), frame, FALSE, FALSE,
			 SPACING);
     gtk_widget_show (frame);

     /* master vbox */
     frame_vbox = gtk_vbox_new (FALSE, SPACING);
     gtk_container_set_border_width (GTK_CONTAINER (frame_vbox), SPACING);
     gtk_container_add (GTK_CONTAINER (frame), frame_vbox);
     gtk_widget_show (frame_vbox);

     /*------------ROW ONE-------------------*/
     scrollwin = gtk_scrolled_window_new (NULL, NULL);
     gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrollwin),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
     gtk_box_pack_start (GTK_BOX (frame_vbox), scrollwin, FALSE, FALSE, 0);
     gtk_widget_show (scrollwin);

     /* create the upper canvas */
     top_canvas = gnome_canvas_new ( );
     gtk_widget_set_size_request (top_canvas, PIANO_DEFAULT_WIDTH,
				  PIANO_DEFAULT_HEIGHT);
     gnome_canvas_set_scroll_region (GNOME_CANVAS (top_canvas), 0, 0,
				     PIANO_DEFAULT_WIDTH,
				     PIANO_DEFAULT_HEIGHT);
     gtk_widget_show (top_canvas);

     /* create piano canvas item */
     piano =
	  PIANO (gnome_canvas_item_new
		 (gnome_canvas_root (GNOME_CANVAS (top_canvas)),
		  piano_get_type ( ), "x", (double) 0, "y", (double) 0,
		  NULL));
     gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrollwin),
					    top_canvas);

     /*Drag and drop handling, in here because it needs a gtkwidget */
     if (GTK_WIDGET_NO_WINDOW (top_canvas))
     {
	  debug ("Widget has no window \n");
     }

     gtk_drag_dest_set (top_canvas, GTK_DEST_DEFAULT_ALL, targets, 1,
			GDK_ACTION_COPY | GDK_ACTION_MOVE);
     g_signal_connect (G_OBJECT (top_canvas), "drag_data_received",
		       G_CALLBACK (piano_dnd), (gpointer) piano);

     /****************************END MAIN WINDOW****************************************/
     gtk_widget_show (left_vbox);
     gtk_widget_show (right_vbox);
     gtk_widget_show (master_hbox);
     gtk_widget_show (window_vbox);
     gtk_widget_show (window);

     /* initialize children */
     sample_editor_init     (window);
     audio_settings_init    (window);
     filter_settings_init   (window);
     playmode_settings_init (window);
     param_settings_init    (window);
     porta_settings_init    (window);
     
     /* initialize */
     update_interface (0);

     /* success */
     return 0;
}

static int current_patch = -1;

/* make all patch related controls sensitive or insensitive */
static void set_sensitive (int val)
{
     gtk_widget_set_sensitive (chan_spin, val);
     gtk_widget_set_sensitive (file_button, val);
     gtk_widget_set_sensitive (note_spin, val);
     gtk_widget_set_sensitive (menu_file_save_bank, val);
     gtk_widget_set_sensitive (menu_file_save_bank_as, val);
     gtk_widget_set_sensitive (play_button, val);
     gtk_widget_set_sensitive (vol_spin, val);
     gtk_widget_set_sensitive (pan_spin, val);
     gtk_widget_set_sensitive (pitch_spin, val);     
     gtk_widget_set_sensitive (pitch_steps_spin, val);
     gtk_widget_set_sensitive (cut_spin, val);
     gtk_widget_set_sensitive (cut_by_spin, val);
     gtk_widget_set_sensitive (range_check, val);
     gtk_widget_set_sensitive (range_low_spin, val);
     gtk_widget_set_sensitive (range_up_spin, val);
}

/* sets current patch and updates widgets accordingly */
static void set_current_patch (int id)
{
     gdouble chan, vol, note, range, lower_note,
	  upper_note, pan, cut, cut_by, frames, pitch;
     int pitch_steps;
     char *sample_name;
     gchar *tmps;

     if (id < 0)
     {
	  debug ("No active patches\n");
	  set_sensitive (FALSE);
	  current_patch = -1;
	  waveform_set_patch (WAVEFORM (waveform), -1);
	  return;
     }

     chan        = (gdouble) patch_get_channel     (id);
     note        = (gdouble) patch_get_note        (id);
     range       = (gdouble) patch_get_range       (id);
     lower_note  = (gdouble) patch_get_lower_note  (id);
     upper_note  = (gdouble) patch_get_upper_note  (id);
     vol         = (gdouble) patch_get_volume      (id);
     pan         = (gdouble) patch_get_panning     (id);
     pitch       = (gdouble) patch_get_pitch       (id);
     pitch_steps =           patch_get_pitch_steps (id);
     cut         = (gdouble) patch_get_cut         (id);
     cut_by      = (gdouble) patch_get_cut_by      (id);
     frames      = (gdouble) patch_get_frames      (id);
     sample_name =           patch_get_sample_name (id);

     debug ("Changing current patch to: %d, index: %d\n", id,
	    patch_get_display_index (id));
     if (current_patch < 0)
	  set_sensitive (TRUE);
     current_patch = id;

     gtk_spin_button_set_value (GTK_SPIN_BUTTON (chan_spin), chan + 1);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (note_spin), note);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (range_low_spin),
				lower_note);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (range_up_spin),
				upper_note);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (vol_spin), vol * 100);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (pan_spin), pan * 100);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (pitch_spin), pitch);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (pitch_steps_spin), pitch_steps);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (cut_spin), cut);
     gtk_spin_button_set_value (GTK_SPIN_BUTTON (cut_by_spin), cut_by);

     if (*sample_name == '\0')
     {
	  gtk_label_set_text (GTK_LABEL (file_label), "Load Sample");
     }
     else
     {
	  gtk_label_set_text (GTK_LABEL (file_label), tmps =
			      g_path_get_basename (sample_name));
	  g_free (tmps);
     }
     if (sample_name != NULL)
	  g_free (sample_name);

     if (range > 0)
     {
	  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (range_check),
					TRUE);
	  gtk_widget_set_sensitive (range_low_spin, TRUE);
	  gtk_widget_set_sensitive (range_up_spin, TRUE);
     }
     else
     {
	  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (range_check),
					FALSE);
	  gtk_widget_set_sensitive (range_low_spin, FALSE);
	  gtk_widget_set_sensitive (range_up_spin, FALSE);
     }

     /* make pretty pictures */
     waveform_set_patch (WAVEFORM (waveform), id);

     return;
}

int get_current_patch ( )
{
     return current_patch;
}
